#include <ZXMod.h>

/*

 NEXT STEP IN "MultiLayering" FRACTALS (Jun 2017)

 RATIONALE:

 Instead of shifting some imaginary boundaries even along multiple dimensions (as done
 previously), we are simply and smartly enough using the original "codename IFS" technique
 (iterating to closest node) by shifting the distance to be compared before comparison, thus
 reducing the problem of domains to one dimension only (kinda).
 These modules are still far from being the "holy graal" sought, but certainly result much
 more flexible than the previous art and have a much higher potential (for some reason, they
 still do not perform well enough to cope with IFSs using INV() though)

 MAPPING:

 Plain logmap (map-n, implemented as productories for speed reasons) of distances from
 every node plus resetting x,y at every shift of m would be the ideal, but accumulators
 tend to saturate quickly; speed also is not optimal at all.
 Also, it would be STUPID not to make use of any CPM-like technique, given the escape dynamics
 of these systems.
 Imposing a limit (l*= r>LIM?LIM:r) works and can be speed improved with CPM-alike technique
 but tends to destroy the neutrality of escaping areas, making the whole result too much
 saturated and "suffocated" (but CAN be adopted nonetheless, depending on the tastes).
 Properly said CPM CAN be adopted but the exponential base is NOT FIXED; therefore a
 sufficiently high break limit must be used (100000 or so) and the base must then be
 computed dynamically upon breaking as base=r/r_prev. Also, +1 must always be added to
 the accumulators to preserve gradient's uniformity (logs of values close to zero tend to
 "collapse inside" otherwise).

 Therefore (this is my CPM greyscale implementation, color will be discussed afterwards):

 if(r< BREAKON) l*= r; else break;

 then: l=log(l) (always);

 n=it-c; if(n>0) ->
 
 l+= log(r*b^0) + log(r*b^1) + log(r*b^...); (b=r/r_prev);
 = log(r^n) * b^(0+1+..n) )
 = log(r) * n + log(b) * (0+1+...n) =
 = log(r) * n + log(b) * (n-n)*0.5 =
 = n * ( log(r) + log(b) * (n-1) * 0.5)

 Changing for l*=(r+1):

 if(r<BREAKON) l*=(r+1); else break;

 ...

 l+= log(r*b^0 +1) + log(r*b^1 +1) + log(r*b^2 +1) + log(r*b^... +1)

 This can't be simplified unfortunately bc it expands polynomially; but by property
 of Limits we can safely ignore the +1 term in the escape approximation.

 To log-map more than one channel, always by virtue of Limits, we can extrapolate
 the base by just ONE of the n-accumulators. Therefore, treating all n accumulators
 orthogonally can luckily be avoided.

 The resulting implementation (LAST non-commented function) other than preventing
 saturation of accumulators also results slightly faster. Default splitting of the shift
 index m can therefore be as high as 40 and still result useable.

 This is the best I could achieve as for Jun 2017 yet without resorting to random or actual
 3D procedures.

 LATER NOTE: Because I could successfully develop true recursive methods shortly having
 worked at this solution, this solution was never exploded or perfectioned any furtherly

*/
 


#define X1  0
#define Y1  1
#define K1  2

#define X2  1
#define Y2  0
#define K2  2

#define X3  0
#define Y3 -1
#define K3  2

#define X4 -1
#define Y4  0
#define K4  2

#define M_RANGE 1.0  // ACTUAL RANGE WILL BE AS TWICE i.e - to + this range
#define M_STEPS 40

#define M_ADDER ((M_RANGE+M_RANGE)/M_STEPS)

#define BREAKON 100000.0


// Original version, plain productorial logmap (saturates quickly in "extern" areas)

/*

ULONG FUNC(x0,y0,a,b,t,it)
{
 register DOUBLE l1,l2,l3,l4,d1,d2,d3,d4,r,m,x,y;
 register DOUBLE s1,s2,s3,s4;
 register ULONG c,cc;

 s1=s2=s3=s4=0;

 for(cc=0,m=-M_RANGE; cc<M_STEPS; cc++, m+=M_ADDER)
 {
  x=x0;
  y=y0;
  l1=l2=l3=l4=1.0;


 for(c=0; c<it; c++)
 {
  d1=((x-X1)*(x-X1)+(y-Y1)*(y-Y1));
  d2=((x-X2)*(x-X2)+(y-Y2)*(y-Y2));
  d3=((x-X3)*(x-X3)+(y-Y3)*(y-Y3));
  d4=((x-X4)*(x-X4)+(y-Y4)*(y-Y4));

  l1*=(d1+1);
  l2*=(d2+1);
  l3*=(d3+1);
  l4*=(d4+1);

  if((d1+m<d2)&&(d1+m<d3)&&(d1+m<d4))
  {
   x=K1*(x-X1)+X1;
   y=K1*(y-Y1)+Y1;
  }
  else if((d2+m<d1)&&(d2+m<d3)&&(d2+m<d4))
  {
   x=K2*(x-X2)+X2;
   y=K2*(y-Y2)+Y2;
   r=-y; y=x; x=r;
  }
  else if((d3+m<d1)&&(d3+m<d2)&&(d3+m<d4))
  {
   x=K3*(x-X3)+X3;
   y=K3*(y-Y3)+Y3;
   x=-x; y=-y;
  }
  else
  {
   x=K4*(x-X4)+X4;
   y=K4*(y-Y4)+Y4;
   r=y; y=-x; x=r;
  }

  CZ;
  //INV(x,y);
 }

 l1=log(l1); l1*=l1; s1+=(cos(log(l1)*t)+1);
 l2=log(l2); l2*=l2; s2+=(cos(log(l2)*t)+1);
 l3=log(l3); l3*=l3; s3+=(cos(log(l3)*t)+1);
 l4=log(l4); l4*=l4; s4+=(cos(log(l4)*t)+1);

}

t=1.0/M_STEPS;

s1*=t;
s2*=t;
s3*=t;
s4*=t;

 RGBRETURN((UBYTE)(s1*63.75  +s2*63.75),
           (UBYTE)(s2*31.875 +s3*31.875 +s4*63.75),
           (UBYTE)(s1*31.775 +s3*63.75  +s4*31.875));
}


*/

/*
// Experimental function (not used)
// Extracts exponent (strangely enough, frexp() is very slow). Can't be written
// in C because the optimizer renders it completely broken. Can conveniently be used
// where qualitative-logarithmic behaviour is needed but only with large numbers

int qexp(__reg("f1") double n) =
"\tstwu\t1,-32(1)\n"
"\tstfd 1,8(1)\n"
"\taddi\t10,1,8\n"
"\tlwz\t 3,0(10)\n"
"\tlis\t12,32752\n"
"\tand\t10,3,12\n"
"\tsrwi\t3,10,20&31\n"
"\taddi\t10,3,-1022\n"
"\tmr\t3,10\n"
"\taddi\t1,1,32";
*/

// Gray CPM-alike with dynamically computed base

/*
ULONG FUNC(x0,y0,a,b,t,it)
{
 register DOUBLE d1,d2,d3,d4,l,r,m,x,y,s,e;
 register ULONG c,cc;

 s=0;

 for(cc=0,m=-M_RANGE; cc<M_STEPS; cc++, m+=M_ADDER)
 {
  x=x0;
  y=y0;
  l=1.0;


 for(c=0; c<it; c++)
 {
  d1=((x-X1)*(x-X1)+(y-Y1)*(y-Y1));
  d2=((x-X2)*(x-X2)+(y-Y2)*(y-Y2));
  d3=((x-X3)*(x-X3)+(y-Y3)*(y-Y3));
  d4=((x-X4)*(x-X4)+(y-Y4)*(y-Y4));

  r=x*x+y*y;

  if(r>100000) break;

  l*=(r);

  e=r;

  if((d1+m<d2)&&(d1+m<d3)&&(d1+m<d4))
  {
   x=K1*(x-X1)+X1;
   y=K1*(y-Y1)+Y1;
  }
  else if((d2+m<d1)&&(d2+m<d3)&&(d2+m<d4))
  {
   x=K2*(x-X2)+X2;
   y=K2*(y-Y2)+Y2;
   r=-y; y=x; x=r;
  }
  else if((d3+m<d1)&&(d3+m<d2)&&(d3+m<d4))
  {
   x=K3*(x-X3)+X3;
   y=K3*(y-Y3)+Y3;
   x=-x; y=-y;
  }
  else
  {
   x=K4*(x-X4)+X4;
   y=K4*(y-Y4)+Y4;
   r=y; y=-x; x=r;
  }

  CZ;
  //INV(x,y);
 }

 l=log(l);


 if(c<it)
 {
  x=it-c; l+= ( x*log(r) + log(r/e)*(x*x-x)*0.5 );
 }


 l*=l; s+=(cos(log(l)*t)+1);

 }

 t=1.0/M_STEPS;

 s*=t;

 c=(int)(s*127.5+0.5);
 RGBRETURN(c,c,c);

}

*/



ULONG FUNC(x0,y0,a,b,t,it)
{
 register DOUBLE r,m,x,y;
 register DOUBLE d1,d2,d3,d4;
 DOUBLE l1,l2,l3,l4,s1,s2,s3,s4;
 register ULONG c,cc;

 s1=s2=s3=s4=0;

 for(cc=0,m=-M_RANGE; cc<M_STEPS; cc++, m+=M_ADDER)
 {
  x=x0;
  y=y0;
  l1=l2=l3=l4=1.0;
  d1=1.0; // just to be 100% sure

  for(c=0; c<it; c++)
  {
   r=d1;
   
   d1=((x-X1)*(x-X1)+(y-Y1)*(y-Y1));
   d2=((x-X2)*(x-X2)+(y-Y2)*(y-Y2));
   d3=((x-X3)*(x-X3)+(y-Y3)*(y-Y3));
   d4=((x-X4)*(x-X4)+(y-Y4)*(y-Y4));  

   if(d1>BREAKON) break;

   l1*=(d1+1); l2*=(d2+1); l3*=(d3+1); l4*=(d4+1);


   if((d1+m<d2)&&(d1+m<d3)&&(d1+m<d4))
   {
    x=K1*(x-X1)+X1;
    y=K1*(y-Y1)+Y1;
   }
   else if((d2+m<d1)&&(d2+m<d3)&&(d2+m<d4))
   {
    x=K2*(x-X2)+X2;
    y=K2*(y-Y2)+Y2;
    r=-y; y=x; x=r;
   }
   else if((d3+m<d1)&&(d3+m<d2)&&(d3+m<d4))
   {
    x=K3*(x-X3)+X3;
    y=K3*(y-Y3)+Y3;
    x=-x; y=-y;
   }
   else
   {
    x=K4*(x-X4)+X4;
    y=K4*(y-Y4)+Y4;
    r=y; y=-x; x=r;
   }

   CZ;
  } // iters loop

  l1=log(l1); l2=log(l2); l3=log(l3); l4=log(l4);

  if(c<it) // c loop broken by large value, simulate escape sequence quickly
  {
   // Borrow x and y as temp variables
   x=it-c;
   y=log(d1/r)*(x-1.0)*0.5;
   l1+= ( x*( log(d1) + y ) );
   l2+= ( x*( log(d2) + y ) );
   l3+= ( x*( log(d3) + y ) );
   l4+= ( x*( log(d4) + y ) );
  }

  l1*=l1; s1+=(cos(log(l1)*t)+1);
  l2*=l2; s2+=(cos(log(l2)*t)+1);
  l3*=l3; s3+=(cos(log(l3)*t)+1);
  l4*=l4; s4+=(cos(log(l4)*t)+1);


 } // m loop

 // stabilize gradients to M_STEPS at least. Borrow t

 t=1.0/M_STEPS;
 s1*=t; s2*=t; s3*=t; s4*=t;

 RGBRETURN((UBYTE)(s1*63.75  +s2*63.75),
           (UBYTE)(s2*31.875 +s3*31.875 +s4*63.75),
           (UBYTE)(s1*31.775 +s3*63.75  +s4*31.875));

}
